// This file is under LGPLv3 license

//#define EXTRA_CHECKS

enum
{
	edge = 4,
	xdia = 8,
	ydia = 6,
	asize = xdia*ydia,
	xdiad2m1 = (xdia>>1)-1,
	in_linex = 24+2,
	in_liney = 72+2,
};

float8 processx_local(__local float (*restrict in)[in_linex], __global const float *restrict weights, uint nnst, uint qual)
{
	float8 sum = 0, sumsq = 0;
	for (uint j = 0; j < ydia; j++)
	{
		float8 t = (float8)(0, in[0][j], in[1][j], in[2][j], in[3][j], in[4][j], in[5][j], in[6][j]);
		#pragma unroll
		for (uint i = 0; i < xdia; i++)
		{
			t = (float8)(t.s1234, t.s567, in[i+7][j]);
			sum += t;
			sumsq += t*t;
		}
	}

	float8 mstd0, mstd1, mstd2, mstd3 = 0;
	mstd0 = sum/(xdia*ydia);
	mstd1 = sumsq/(xdia*ydia)-mstd0*mstd0;
	mstd2 = mstd1 > FLT_EPSILON ? rsqrt(mstd1) : 0;
	mstd1 = mstd1*mstd2;

	for (int q=0; q<qual; ++q)
	{
		float8 vsum = 0, wsum = 0;
		for (int k = 0; k < nnst; k++)
		{
			float8 sum1 = 0;
			float8 sum2 = 0;
			#pragma unroll 2
			for (uint j = 0; j < ydia; j++)
			{
				float w[16];
				*((float16*)w) = *(__global const float16*)weights;
				weights += 16;
				float8 t = (float8)(0, in[0][j], in[1][j], in[2][j], in[3][j], in[4][j], in[5][j], in[6][j]);
				#pragma unroll
				for (uint i = 0; i < xdia; i++)
				{
					t = (float8)(t.s1234, t.s567, in[i+7][j]);
					sum1 += t*w[i];
					sum2 += t*w[i+8];
				}
			}
			float2 w = *(__global const float2*)weights;
			weights += 4;
#ifdef EXTRA_CHECKS
			sum1 = exp(max(min(sum1*mstd2+w.s0,80.0f),-80.0f));
#else
			sum1 = exp(sum1*mstd2+w.s0);
#endif
			sum2 = sum2*mstd2+w.s1;
			wsum += sum1;
			vsum += sum1*(sum2/(1.0f+fabs(sum2)));
		}
#ifdef EXTRA_CHECKS
		mstd3 += wsum > 1e-10f ? 5.0f*vsum/wsum*mstd1 + mstd0 : mstd0;
#else
		mstd3 += mstd0 + 5.0f*vsum/wsum*mstd1;
#endif
	}
	return mstd3*(1.0f/qual);
}

float8 processy_local(__local float (*restrict in)[in_liney], __global const float *restrict weights, uint nnst, uint qual)
{
	float8 sum = 0, sumsq = 0;
	for (uint j = 0; j < ydia; j++)
	{
#if defined(__GPU__) && defined(__AMD__) && __OPENCL_VERSION__ >= 110
		float8 t = (float8)((*(__local float3*)&in[j][0]).s0012, *(__local float4*)&in[j][3]);
#else
		float8 t = (float8)(0, in[j][0], in[j][1], in[j][2], in[j][3], in[j][4], in[j][5], in[j][6]);
#endif
		#pragma unroll
		for (uint i = 0; i < xdia; i++)
		{
			t = (float8)(t.s1234, t.s567, in[j][i+7]);
			sum += t;
			sumsq += t*t;
		}
	}

	float8 mstd0, mstd1, mstd2, mstd3 = 0;
	mstd0 = sum/(xdia*ydia);
	mstd1 = sumsq/(xdia*ydia)-mstd0*mstd0;
	mstd2 = mstd1 > FLT_EPSILON ? rsqrt(mstd1) : 0;
	mstd1 = mstd1*mstd2;

	for (int q=0; q<qual; ++q)
	{
		float8 vsum = 0, wsum = 0;
		for (int k = 0; k < nnst; k++)
		{
			float8 sum1 = 0;
			float8 sum2 = 0;
			#pragma unroll 2
			for (uint j = 0; j < ydia; j++)
			{
				float w[16];
				*((float16*)w) = *(__global const float16*)weights;
				weights += 16;
#if defined(__GPU__) && defined(__AMD__) && __OPENCL_VERSION__ >= 110
				float8 t = (float8)((*(__local float3*)&in[j][0]).s0012, *(__local float4*)&in[j][3]);
#else
				float8 t = (float8)(0, in[j][0], in[j][1], in[j][2], in[j][3], in[j][4], in[j][5], in[j][6]);
#endif
				#pragma unroll
				for (uint i = 0; i < xdia; i++)
				{
					t = (float8)(t.s1234, t.s567, in[j][i+7]);
					sum1 += t*w[i];
					sum2 += t*w[i+8];
				}
			}
			float2 w = *(__global const float2*)weights;
			weights += 4;
#ifdef EXTRA_CHECKS
			sum1 = exp(max(min(sum1*mstd2+w.s0,80.0f),-80.0f));
#else
			sum1 = exp(sum1*mstd2+w.s0);
#endif
			sum2 = sum2*mstd2+w.s1;
			wsum += sum1;
			vsum += sum1*(sum2/(1.0f+fabs(sum2)));
		}
#ifdef EXTRA_CHECKS
		mstd3 += wsum > 1e-10f ? 5.0f*vsum/wsum*mstd1 + mstd0 : mstd0;
#else
		mstd3 += mstd0 + 5.0f*vsum/wsum*mstd1;
#endif
	}
	return mstd3*(1.0f/qual);
}

// 2D: 64x1 group, any|16*4 x any|32/32 grid
__kernel __attribute__((reqd_work_group_size(64, 1, 1)))
void processx(__global const float *restrict weights, __global uchar *restrict buf, uint stride, uint dx, uint nnst, uint qual, uint offset_in, uint offset_out)
{
	__local float input[39][in_linex];
	uint x, y;

	y = get_local_id(0) / 3;
	x = get_local_id(0) - mul24(y, 3u);
	__global uchar *buf_in = buf + mad24(get_group_id(1)*32 + y + edge-xdiad2m1, stride, get_group_id(0)*16 + x*8 + offset_in);
#if defined(__GPU__) && defined(__AMD__)
	((__local float8*)input[y])[x] = convert_float8(*((__global const uchar8*)buf_in));
	if (y < 39-21)
		((__local float8*)input[y+21])[x] = convert_float8(*((__global const uchar8*)&buf_in[stride*21]));
#else
	vstore8(convert_float8(*((__global const uchar8*)buf_in)), x, input[y]);
	if (y < 39-21)
		vstore8(convert_float8(*((__global const uchar8*)&buf_in[stride*21])), x, input[y+21]);
#endif
	barrier(CLK_LOCAL_MEM_FENCE);

	x = get_local_id(0) & 15;
	y = get_local_id(0) / 16;
	__local float (*in)[in_linex] = (__local float(*)[in_linex])&input[y*8][x + edge-ydia/2 + dx];
	float8 mstd3 = processx_local(in, weights, nnst, qual);
	float8 s;
	float16 df;
	s = (float8)(in[3][3-dx], in[4][3-dx], in[5][3-dx], in[6][3-dx], in[7][3-dx], in[8][3-dx], in[9][3-dx], in[10][3-dx]);
	if (dx)
	{
		df.even = s;
		df.odd = mstd3;
	}
	else
	{
		df.even = mstd3;
		df.odd = s;
	}
	uchar4 d[4] = {convert_uchar4_sat_rte(df.s0123), convert_uchar4_sat_rte(df.s4567), convert_uchar4_sat_rte(df.s89ab), convert_uchar4_sat_rte(df.scdef)};
	barrier(CLK_LOCAL_MEM_FENCE);

	__local uchar4 (*out)[16] = (__local uchar4(*)[16])input;
	out[y*4  ][x] = d[0];
	out[y*4+1][x] = d[1];
	out[y*4+2][x] = d[2];
	out[y*4+3][x] = d[3];
	barrier(CLK_LOCAL_MEM_FENCE);

	x = get_local_id(0) & 3;
	y = get_local_id(0) / 4;
	d[0] = out[y][x*4  ];
	d[1] = out[y][x*4+1];
	d[2] = out[y][x*4+2];
	d[3] = out[y][x*4+3];

	uchar8 d0 = (uchar8)(d[0].lo, d[1].lo, d[2].lo, d[3].lo);
	uchar8 d1 = (uchar8)(d[0].hi, d[1].hi, d[2].hi, d[3].hi);

	buf += mad24(get_group_id(1)*32+y*2, stride, get_group_id(0)*32 + x*8 + offset_out);
	*((__global uchar8*)buf) = d0;
	*((__global uchar8*)(buf + stride)) = d1;
}

// 2D: 8x8 group, any|64/8 x any|8 grid
__kernel __attribute__((reqd_work_group_size(8, 8, 1)))
void processy(__global const float *restrict weights, __global uchar *restrict buf, uint stride, uint dy, uint nnst, uint qual)
{
	__local float input[13][in_liney];

	buf += mad24(get_group_id(1)*8+5-ydia/2, stride, mad24(stride/2, dy, get_group_id(0)*64));
	__global uchar *buf_out = buf + mad24(get_local_id(1)+edge-2, stride, get_local_id(0)*8 + edge);

	uint xy = get_local_id(1)*8 + get_local_id(0);
	uint y = xy / 9;
	uint x = xy - mul24(y, 9u);
	buf += mad24(y, stride, x*8 - stride/2);

#if defined(__GPU__) && defined(__AMD__)
	if (xy != 63)
	{
		((__local float8*)input[y])[x] = convert_float8(*((__global const uchar8*)buf));
		if (y != 6)
			((__local float8*)input[y+7])[x] = convert_float8(*((__global const uchar8*)&buf[stride*7]));
	}
#else
	//if (xy != 63)
	{
		vstore8(convert_float8(*((__global const uchar8*)buf)), x, input[y]);
		if (y < 6)
			vstore8(convert_float8(*((__global const uchar8*)&buf[stride*7])), x, input[y+7]);
	}
#endif
	barrier(CLK_LOCAL_MEM_FENCE);

	float8 mstd3 = processy_local((__local float(*)[in_liney])&input[get_local_id(1)][get_local_id(0)*8 + edge-xdiad2m1], weights, nnst, qual);

	*((__global uchar8*)buf_out) = convert_uchar8_sat_rte(mstd3);
}

// 1D: any
__kernel
void padx(__global uchar *restrict buf, uint stride, uint width, uint height, uint offset_in)
{
	uint i = mad24(get_group_id(0), get_local_size(0), get_local_id(0));
	if (i < height)
	{
		buf += mad24(i, stride, offset_in);
		__global uint *buf2 = (__global uint*)(buf + (width - 1 & ~3));
		uchar4 s = (*((__global uchar16*)buf)).s8765;
		i = (width - 1 & 3)*8;
		uint a = buf2[1];
		uint c = as_uint(as_uchar4((uint)(upsample(a, buf2[0]) >> i)).s3210);
		i = 24-i;
		*((__global uchar4*)buf) = s;
		buf2[1] = upsample(c, a << i) >> i;
		buf2[2] = c >> i;
	}
}

// 2D: any x 8 group, any|16/16 x 8 grid
__kernel
void pady(__global uchar *restrict buf, uint stride, uint width, uint offset_top, uint offset_bottom)
{
	uint x = mad24(get_group_id(0), get_local_size(0)*16, get_local_id(0)*16);
	if (x < width)
	{
		uint y = get_local_id(1);
		bool z = y >= 4;
		int i = z ? -1 : 4;
		y &= 3;
		buf += mad24(y, stride, (z ? offset_bottom : offset_top) + x);
		int offset = mul24(i - (int)y, (int)stride*2);
		*((__global uchar16*)buf) = *((__global uchar16*)(buf + offset));
	}
}
